# coding=utf-8

import tensorflow as tf
# from tf_geometric.sparse.sparse_adj import SparseAdj
from tf_sparse import SparseMatrix

from tf_geometric.nn.conv.gcn import gcn_norm_adj
import tf_sparse as tfs

def sgc(x, edge_index, edge_weight, k, kernel, bias=None, activation=None, renorm=True, improved=False, cache=None):
    """
    Functional API for Simple Graph Convolution (SGC).

    :param x: Tensor, shape: [num_nodes, num_features], node features
    :param edge_index: Tensor, shape: [2, num_edges], edge information
    :param edge_weight: Tensor or None, shape: [num_edges]
    :param k: Number of hops.(default: :obj:`1`)
    :param kernel: Tensor, shape: [num_features, num_output_features], weight.
    :param bias: Tensor, shape: [num_output_features], bias.
    :param activation: Activation function to use.
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
    :return: Updated node features (x), shape: [num_nodes, num_features]
    """
    num_nodes = tfs.shape(x)[0]
    sparse_adj = SparseMatrix(edge_index, edge_weight, [num_nodes, num_nodes])
    normed_sparse_adj = gcn_norm_adj(sparse_adj, renorm=renorm, improved=improved, cache=cache)

    # SparseTensor is usually used for one-hot node features (For example, feature-less nodes.)
    if isinstance(x, tf.sparse.SparseTensor):
        h = tf.sparse.sparse_dense_matmul(x, kernel)
    else:
        h = x @ kernel

    for _ in range(k):
        h = normed_sparse_adj @ h

    # updated_edge_index, normed_edge_weight = gcn_norm_edge(edge_index, x.shape[0], edge_weight,
    #                                                        renorm, improved, cache)
    #
    # h = x
    # for _ in range(k):
    #     h = aggregate_neighbors(
    #         h,
    #         updated_edge_index,
    #         normed_edge_weight,
    #         gcn_mapper,
    #         sum_reducer,
    #         identity_updater
    #     )

    # h = h @ kernel

    if bias is not None:
        h += bias

    if activation is not None:
        h = activation(h)

    return h




# coding=utf-8
import tensorflow as tf
# from tf_geometric.sparse.sparse_adj import SparseAdj
from tf_sparse import SparseMatrix

from tf_geometric.nn.conv.gcn import gcn_norm_adj
import tf_sparse as tfs


def tagcn(x, edge_index, edge_weight, k, kernel, bias=None, activation=None, renorm=False, improved=False, cache=None):
    """
    Functional API for Topology Adaptive Graph Convolutional Network (TAGCN).

    :param x: Tensor, shape: [num_nodes, num_features], node features.
    :param edge_index: Tensor, shape: [2, num_edges], edge information.
    :param edge_weight: Tensor or None, shape: [num_edges].
    :param k: Number of hops.(default: :obj:`3`)
    :param kernel: Tensor, shape: [num_features, num_output_features], weight.
    :param bias: Tensor, shape: [num_output_features], bias.
    :param activation: Activation function to use.
    :param renorm: Whether use renormalization trick (https://arxiv.org/pdf/1609.02907.pdf).
    :param improved: Whether use improved GCN or not.
    :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
    :return: Updated node features (x), shape: [num_nodes, num_output_features]
    """

    num_nodes = tfs.shape(x)[0]

    sparse_adj = SparseMatrix(edge_index, edge_weight, [num_nodes, num_nodes])
    normed_sparse_adj = gcn_norm_adj(sparse_adj, renorm=renorm, improved=improved, cache=cache)

    if isinstance(x, tf.sparse.SparseTensor):
        x = tf.sparse.to_dense(x)
    elif isinstance(x, tfs.SparseMatrix):
        x = x.to_dense()

    xs = [x]
    for _ in range(k):
        h = normed_sparse_adj @ xs[-1]
        xs.append(h)

    h = tf.concat(xs, axis=-1)

    out = h @ kernel
    if bias is not None:
        out += bias

    if activation is not None:
        out = activation(out)

    return out



# coding=utf-8

from tf_geometric.nn.conv.gcn import gcn_build_cache_for_graph, gcn_build_cache_by_adj
from tf_geometric.nn.conv.appnp import appnp
import tensorflow as tf
import warnings


class APPNP(tf.keras.Model):

    def build(self, input_shapes):
        x_shape = input_shapes[0]
        last_units = x_shape[-1]

        for i, units in enumerate(self.units_list):
            kernel_name = "kernel_{}".format(i)
            kernel = self.add_weight(kernel_name, shape=[last_units, units],
                                     initializer="glorot_uniform", regularizer=self.kernel_regularizer)

            bias_name = "bias_{}".format(i)
            bias = self.add_weight(bias_name, shape=[units],
                                   initializer="zeros", regularizer=self.bias_regularizer)
            last_units = units

            self.kernels.append(kernel)
            self.biases.append(bias)

            # setattr(self, kernel_name, kernel)
            # setattr(self, bias_name, bias)

    def __init__(self, units_list,
                 dense_activation=tf.nn.relu, activation=None,
                 k=10, alpha=0.1,
                 dense_drop_rate=0.0, last_dense_drop_rate=0.0, edge_drop_rate=0.0,
                 kernel_regularizer=None, bias_regularizer=None, *args, **kwargs):
        """

        :param units_list: List of Positive integers consisting of dimensionality of the output space of each dense layer.
        :param dense_activation: Activation function to use for the dense layers,
            except for the last dense layer, which will not be activated.
        :param activation: Activation function to use for the output.
        :param k: Number of propagation power iterations.
        :param alpha: Teleport Probability.
        :param dense_drop_rate: Dropout rate for the output of every dense layer (except the last one).
        :param last_dense_drop_rate: Dropout rate for the output of the last dense layer.
            last_dense_drop_rate is usually set to 0.0 for classification tasks.
        :param edge_drop_rate: Dropout rate for the edges/adj used for propagation.
        :param kernel_regularizer: Regularizer function applied to the `kernel` weights matrices.
        :param bias_regularizer: Regularizer function applied to the bias vectors.
        """

        super().__init__(*args, **kwargs)
        self.units_list = units_list
        self.dense_activation = dense_activation
        self.activation = activation
        self.k = k
        self.alpha = alpha

        self.dense_drop_rate = dense_drop_rate
        self.last_dense_drop_rate = last_dense_drop_rate
        self.edge_drop_rate = edge_drop_rate

        self.kernel_regularizer = kernel_regularizer
        self.bias_regularizer = bias_regularizer

        self.kernels = []
        self.biases = []


    def build_cache_by_adj(self, sparse_adj, override=False, cache=None):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        return gcn_build_cache_by_adj(sparse_adj, override=override, cache=cache)


    def build_cache_for_graph(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None
        """
        gcn_build_cache_for_graph(graph, override=override)

    def cache_normed_edge(self, graph, override=False):
        """
        Manually compute the normed edge based on this layer's GCN normalization configuration (self.renorm and self.improved) and put it in graph.cache.
        If the normed edge already exists in graph.cache and the override parameter is False, this method will do nothing.

        :param graph: tfg.Graph, the input graph.
        :param override: Whether to override existing cached normed edge.
        :return: None

        .. deprecated:: 0.0.56
            Use ``build_cache_for_graph`` instead.
        """
        warnings.warn(
            "'APPNP.cache_normed_edge(graph, override)' is deprecated, use 'APPNP.build_cache_for_graph(graph, override)' instead",
            DeprecationWarning)
        return self.build_cache_for_graph(graph, override=override)

    def call(self, inputs, cache=None, training=None, mask=None):
        """

        :param inputs: List of graph info: [x, edge_index, edge_weight]
        :param cache: A dict for caching A' for GCN. Different graph should not share the same cache dict.
        :return: Updated node features (x), shape: [num_nodes, units]
        """

        if len(inputs) == 3:
            x, edge_index, edge_weight = inputs
        else:
            x, edge_index = inputs
            edge_weight = None

        return appnp(x, edge_index, edge_weight, self.kernels, self.biases,
                     dense_activation=self.dense_activation, activation=self.activation,
                     k=self.k, alpha=self.alpha,
                     dense_drop_rate=self.dense_drop_rate,
                     last_dense_drop_rate=self.last_dense_drop_rate,
                     edge_drop_rate=self.edge_drop_rate,
                     cache=cache, training=training)



# coding=utf-8

import tf_geometric as tfg
import numpy as np


graph0 = tfg.HeteroGraph(
    x_dict={
        "paper": np.random.randn(2, 5),
        "author": np.random.randn(4, 5)
    },
    edge_index_dict={
        ("paper", "r.write", "author"): np.array([
            [0, 0, 1, 1],
            [0, 2, 1, 3]
        ])
    },
    y_dict={
        "paper": np.array([0, 1])
    }
)



graph1 = tfg.HeteroGraph(
    x_dict={
        "paper": np.random.randn(3, 5),
        "author": np.random.randn(5, 5)
    },
    edge_index_dict={
        ("paper", "r.write", "author"): np.array([
            [0, 1, 1, 1, 2, 2],
            [0, 0, 1, 2, 3, 4]
        ])
    },
    y_dict={
        "paper": np.array([1, 1, 0])
    }
)

print(graph0)
print(graph1)

batch_graph = tfg.HeteroBatchGraph.from_graphs([graph0, graph1])

print(batch_graph)
print(batch_graph.edge_index_dict)
print(batch_graph.node_graph_index_dict)
print(batch_graph.edge_graph_index_dict)

batch_graph = batch_graph.convert_data_to_tensor()
print(batch_graph.edge_index_dict)

print(batch_graph.y_dict)

batch_graph.graphs = None

print(batch_graph.num_graphs)


# coding=utf-8
import os.path
import json
from collections import defaultdict
import numpy as np
from tf_geometric.data.graph import HeteroGraph
from tf_geometric.data.dataset import DownloadableDataset


class HGBDataset(DownloadableDataset):

    def __init__(self, dataset_name, dataset_root_path=None):
        """

        :param dataset_name: "hgb_acm" | "hgb_dblp" | "hgb_freebase" | "hgb_imdb"
        :param dataset_root_path:
        """
        self.sub_dataset_name = dataset_name.split("_")[1]

        super().__init__(dataset_name=dataset_name,
                         download_urls=[
                             "https://github.com/CrawlScript/gnn_datasets/raw/master/hgb/{}.zip".format(self.sub_dataset_name)
                         ],
                         download_file_name="{}.zip".format(self.sub_dataset_name),
                         cache_name=None,
                         dataset_root_path=dataset_root_path,
                         )

    # https://github.com/pyg-team/pytorch_geometric/blob/master/torch_geometric/datasets/hgb_dataset.py
    def process(self):
        # data = HeteroData()

        # node_types = {0: 'paper', 1, 'author', ...}
        # edge_types = {0: ('paper', 'cite', 'paper'), ...}

        data_dir = os.path.join(self.raw_root_path, self.sub_dataset_name)

        if self.sub_dataset_name in ['acm', 'dblp', 'imdb']:

            # with open(self.raw_paths[0], 'r') as f:  # `info.dat`
            with open(os.path.join(data_dir, "info.dat"), 'r') as f:  # `info.dat`
                info = json.load(f)
            n_types = info['node.dat']['node type']
            n_types = {int(k): v for k, v in n_types.items()}
            e_types = info['link.dat']['link type']
            e_types = {int(k): tuple(v.values()) for k, v in e_types.items()}
            for key, (src, dst, rel) in e_types.items():
                src, dst = n_types[int(src)], n_types[int(dst)]
                rel = rel.split('-')[1]
                rel = rel if rel != dst and rel[1:] != dst else 'to'
                e_types[key] = (src, rel, dst)
            num_classes = len(info['label.dat']['node type']['0'])
        elif self.sub_dataset_name in ['freebase']:
            # with open(self.raw_paths[0], 'r') as f:  # `info.dat`
            with open(os.path.join(data_dir, "info.dat"), 'r') as f:  # `info.dat`
                info = f.read().split('\n')
            start = info.index('TYPE\tMEANING') + 1
            end = info[start:].index('')
            n_types = [v.split('\t\t') for v in info[start:start + end]]
            n_types = {int(k): v.lower() for k, v in n_types}

            e_types = {}
            start = info.index('LINK\tSTART\tEND\tMEANING') + 1
            end = info[start:].index('')
            for key, row in enumerate(info[start:start + end]):
                row = row.split('\t')[1:]
                src, dst, rel = [v for v in row if v != '']
                src, dst = n_types[int(src)], n_types[int(dst)]
                rel = rel.split('-')[1]
                e_types[key] = (src, rel, dst)
        else:  # Link prediction:
            raise NotImplementedError

        # Extract node information:
        mapping_dict = {}  # Maps global node indices to local ones.
        x_dict = defaultdict(list)
        num_nodes_dict = defaultdict(lambda: 0)

        # with open(self.raw_paths[1], 'r') as f:  # `node.dat`
        with open(os.path.join(data_dir, "node.dat"), 'r') as f:  # `node.dat
            xs = [v.split('\t') for v in f.read().split('\n')[:-1]]
        for x in xs:
            n_id, n_type = int(x[0]), n_types[int(x[2])]
            mapping_dict[n_id] = num_nodes_dict[n_type]
            num_nodes_dict[n_type] += 1
            if len(x) >= 4:  # Extract features (in case they are given).
                x_dict[n_type].append([float(v) for v in x[3].split(',')])
            else:
                x_dict[n_type].append([np.inf])


        x_dict = {ntype: np.array(x, dtype=np.float64) for ntype, x in x_dict.items()}


        # for n_type in n_types.values():
        #     if len(x_dict[n_type]) == 0:
        #         data[n_type].num_nodes = num_nodes_dict[n_type]
        #     else:
        #         data[n_type].x = torch.tensor(x_dict[n_type])

        # edge_index_dict = defaultdict(list)
        # edge_weight_dict = defaultdict(list)

        edge_dict = defaultdict(list)
        edge_weight_dict = defaultdict(list)

        # with open(self.raw_paths[2], 'r') as f:  # `link.dat`
        with open(os.path.join(data_dir, "link.dat"), 'r') as f:  # `link.dat`
            edges = [v.split('\t') for v in f.read().split('\n')[:-1]]
        for src, dst, rel, weight in edges:
            e_type = e_types[int(rel)]
            src, dst = mapping_dict[int(src)], mapping_dict[int(dst)]
            edge_dict[e_type].append([src, dst])
            edge_weight_dict[e_type].append(float(weight))

        edge_index_dict = {e_type: np.array(edges, dtype=np.int64).T for e_type, edges in edge_dict.items()}
        edge_weight_dict = {e_type: np.array(edge_weight, dtype=np.float64)
                            for e_type, edge_weight in edge_weight_dict.items()
                            if not np.allclose(edge_weight, np.ones_like(edge_weight))}

        # for e_type in e_types.values():
        #     edge_index = torch.tensor(edge_index_dict[e_type])
        #     edge_weight = torch.tensor(edge_weight_dict[e_type])
        #     data[e_type].edge_index = edge_index.t().contiguous()
        #     # Only add "weighted" edgel to the graph:
        #     if not torch.allclose(edge_weight, torch.ones_like(edge_weight)):
        #         data[e_type].edge_weight = edge_weight

        y_dict = {}
        train_mask_dict = {}
        test_mask_dict = {}


        # with open(self.raw_paths[3], 'r') as f:  # `label.dat`
        with open(os.path.join(data_dir, "label.dat"), 'r') as f:  # `label.dat`
            train_ys = [v.split('\t') for v in f.read().split('\n')[:-1]]
        # with open(self.raw_paths[4], 'r') as f:  # `label.dat.test`
        with open(os.path.join(data_dir, "label.dat.test"), 'r') as f:  # `label.dat.test`
            test_ys = [v.split('\t') for v in f.read().split('\n')[:-1]]

        for y in train_ys:
            n_id, n_type = mapping_dict[int(y[0])], n_types[int(y[2])]

            if n_type not in y_dict:

                num_nodes = x_dict[n_type].shape[0]# if n_type in x_dict else num_nodes_dict[n_type]

                if self.sub_dataset_name in ['imdb']:  # multi-label
                    y_dict[n_type] = np.zeros([num_nodes, num_classes], dtype=np.int64)
                else:
                    y_dict[n_type] = np.full([num_nodes], -1, dtype=np.int64)

                train_mask_dict[n_type] = np.zeros(num_nodes).astype(np.bool)
                test_mask_dict[n_type] = np.zeros(num_nodes).astype(np.bool)


            # if not hasattr(data[n_type], 'y'):
            #     num_nodes = data[n_type].num_nodes
            #     if self.name in ['imdb']:  # multi-label
            #         data[n_type].y = torch.zeros((num_nodes, num_classes))
            #     else:
            #         data[n_type].y = torch.full((num_nodes,), -1).long()
            #     data[n_type].train_mask = torch.zeros(num_nodes).bool()
            #     data[n_type].test_mask = torch.zeros(num_nodes).bool()

            if y_dict[n_type].ndim > 1:  # multi-label
                for v in y[3].split(','):
                    y_dict[n_type][n_id, int(v)] = 1
            else:
                y_dict[n_type][n_id] = int(y[3])
            train_mask_dict[n_type][n_id] = True

            # if data[n_type].y.dim() > 1:  # multi-label
            #     for v in y[3].split(','):
            #         data[n_type].y[n_id, int(v)] = 1
            # else:
            #     data[n_type].y[n_id] = int(y[3])
            # data[n_type].train_mask[n_id] = True

        for y in test_ys:
            n_id, n_type = mapping_dict[int(y[0])], n_types[int(y[2])]
            if y_dict[n_type].ndim > 1:  # multi-label
                for v in y[3].split(','):
                    y_dict[n_type][n_id, int(v)] = 1
            else:
                y_dict[n_type][n_id] = int(y[3])
            test_mask_dict[n_type][n_id] = True


        # for y in test_ys:
        #     n_id, n_type = mapping_dict[int(y[0])], n_types[int(y[2])]
        #     if data[n_type].y.dim() > 1:  # multi-label
        #         for v in y[3].split(','):
        #             data[n_type].y[n_id, int(v)] = 1
        #     else:
        #         data[n_type].y[n_id] = int(y[3])
        #     data[n_type].test_mask[n_id] = True


        hetero_graph = HeteroGraph(x_dict=x_dict,
                                       edge_index_dict=edge_index_dict,
                                       y_dict=y_dict,
                                       edge_weight_dict=edge_weight_dict)

        return hetero_graph, train_mask_dict, test_mask_dict


class HGBACMDataset(HGBDataset):
    def __init__(self, dataset_root_path=None):
        super().__init__("hgb_acm", dataset_root_path)


class HGBDBLPDataset(HGBDataset):
    def __init__(self, dataset_root_path=None):
        super().__init__("hgb_dblp", dataset_root_path)


class HGBFreebaseDataset(HGBDataset):
    def __init__(self, dataset_root_path=None):
        super().__init__("hgb_freebase", dataset_root_path)


class HGBIMDBDataset(HGBDataset):
    def __init__(self, dataset_root_path=None):
        super().__init__("hgb_imdb", dataset_root_path)



# coding=utf-8
import os

from tensorflow.python.keras.utils.data_utils import _extract_archive
from tf_geometric.utils.data_utils import download_file, load_cache, save_cache
from shutil import copy

DEFAULT_DATASETS_ROOT = "data"


def get_dataset_root_path(dataset_root_path=None, dataset_name=None,
                          datasets_root_path=DEFAULT_DATASETS_ROOT, mkdir=False):
    if dataset_root_path is None:
        dataset_root_path = os.path.join(datasets_root_path, dataset_name)
    dataset_root_path = os.path.abspath(dataset_root_path)

    if mkdir:
        os.makedirs(dataset_root_path, exist_ok=True)
    return dataset_root_path


class Dataset(object):
    pass


class DownloadableDataset(object):

    def __init__(self,
                 dataset_name,
                 download_urls=None,
                 download_file_name=None,
                 cache_name="cache.p",
                 dataset_root_path=None
                 ):
        self.dataset_name = dataset_name
        self.dataset_root_path = get_dataset_root_path(dataset_root_path, dataset_name)
        self.download_urls = download_urls
        self.download_file_name = download_file_name

        self.download_root_path = os.path.join(self.dataset_root_path, "download")
        self.raw_root_path = os.path.join(self.dataset_root_path, "raw")
        self.processed_root_path = os.path.join(self.dataset_root_path, "processed")

        if download_urls is not None:
            if download_file_name is None:
                download_file_name = "{}.zip".format(dataset_name)
            self.download_file_path = os.path.join(self.download_root_path, download_file_name)
        else:
            self.download_file_path = None

        self.cache_path = None if cache_name is None else os.path.join(self.processed_root_path, cache_name)

        self.build_dirs()

    @property
    def cache_enabled(self):
        return self.cache_path is not None

    def build_dirs(self):
        os.makedirs(self.download_root_path, exist_ok=True)
        os.makedirs(self.raw_root_path, exist_ok=True)
        os.makedirs(self.processed_root_path, exist_ok=True)

    def download(self):
        download_file(self.download_file_path, self.download_urls)

    def extract_raw(self):
        if len(os.listdir(self.raw_root_path)) == 0:
            if self.download_file_path.endswith(".npz"):
                copy(self.download_file_path, os.path.join(self.raw_root_path, self.download_file_name))
            else:
                _extract_archive(self.download_file_path, self.raw_root_path, archive_format="auto")
        else:
            print("raw data exists: {}, ignore".format(self.raw_root_path))

    def process(self):
        pass

    def load_data(self):
        if self.cache_enabled and os.path.exists(self.cache_path):
            print("cache file exists: {}, read cache".format(self.cache_path))
            return load_cache(self.cache_path)

        if self.download_urls is not None:
            self.download()
            self.extract_raw()
        else:
            print("downloading and extraction are ignored due to None download_urls")

        processed = self.process()

        if self.cache_enabled:
            print("save processed data to cache: ", self.cache_path)
            save_cache(processed, self.cache_path)

        return processed




















